import asyncio

import aiohttp
from pysmartthings import SmartThings

print(
    "If you do not have an API key, generate"
    " one here: https://account.smartthings.com/tokens"
)

key = input("\nPlease enter the smartthings API key:\n")
device_id = input("Please enter the device network ID (blank for ALL):\n")
print()


async def switch():
    async with aiohttp.ClientSession() as session:
        api = SmartThings(session, key)
        devices = await api.devices()
        if len(devices) == 0:
            print("no devices for that API key")
        for device in devices:
            print("found device: ", device.device_id, device.name, device.label)
            if device.device_id == device_id or device_id == "":
                await device.status.refresh()
                print(" device status is :" + device.status.values.get("switch", ""))


asyncio.run(switch())



"""
cisco_ios_poe_switch.py
--------------

Classes to represent the configuration and functionality for devices
that can be controlled via Cisco IOS shell commands.
"""

from dataclasses import dataclass
from typing import Annotated as A
from typing import Literal

from netmiko import ConnectHandler

from maaspower.maas_globals import desc
from maaspower.maasconfig import SwitchDevice


@dataclass(kw_only=True)
class CiscoIOSPOESwitch(SwitchDevice):
    """A device controlled via a Cisco IOS PoE switch"""

    name: A[
        str, desc("Logical device name for server connected to the Cisco switch port")
    ]
    ip_address: A[str, desc("IP address of the Cisco IOS PoE switch")]
    username: A[str, desc("Username for the Cisco IOS PoE switch")]
    password: A[str, desc("Password for the Cisco IOS PoE switch")]
    enable_password: A[
        str,
        desc("Enable/secret password for escalating privileges on the Cisco switch"),
    ] = ""
    port_selection_string: A[
        str,
        desc(
            """Port selection for the target device connected to the switch in
            Cisco format"""
        ),
    ]
    port_poe_watts: A[
        int, desc("Power budget (in watts) for the target Cisco switch port")
    ] = -1

    type: Literal["CiscoIOSPOESwitch"] = "CiscoIOSPOESwitch"

    def __post_init__(self):
        """
        Create the 'device' variable, containing properties for establishing \n
        a connection via netmiko.
        """

        super().__post_init__()
        self.device = {
            "device_type": "cisco_ios",
            "host": self.ip_address,
            "username": self.username,
            "password": self.password,
            "secret": self.enable_password,
            "fast_cli": False,
            "allow_auto_change": False,
        }

    def turn_on(self) -> None:
        self._change_power_state(True)

    def turn_off(self) -> None:
        self._change_power_state(False)

    def _change_power_state(self, state: bool):
        """
        Takes a boolean representation of the desired port power state and \n
        affects the desired state on the switch via netmiko.

        Args:
            state: A boolean representation of the power state of the switch port.
        """

        if not state:
            power_cmdline = "power inline never"
        elif state:
            if self.port_poe_watts > 0:
                power_cmdline = f"power inline static max {self.port_poe_watts*1000}"
            else:
                power_cmdline = "power inline auto"
        command_set = [f"interface {self.port_selection_string}", power_cmdline]
        try:
            with ConnectHandler(**self.device) as cisco_conn:
                cisco_conn.enable(check_state=False)
                cisco_conn.send_config_set(command_set)
                cisco_conn.disconnect()
        except Exception as e:
            print(
                f"""Failed to turn device power f{'on' if state else 'off'} for
                port {self.port_selection_string}. Reason: {e}"""
            )
            return "error"

    def query_state(self) -> str:
        """
        Requests the configuration of the desired switch port and parses it to \n
        determine whether the power state is on or off.
        """

        try:
            with ConnectHandler(**self.device) as cisco_conn:
                cisco_conn.enable(check_state=False)
                output = cisco_conn.send_command(
                    f"show running-config interface {self.port_selection_string}"
                )
                cisco_conn.disconnect()
                if "power inline never" in str(output).lower():
                    return "off"
                else:
                    return "on"
        except Exception as e:
            print(f"Failed to query device port state with reason: {e}")
            return "error"



import json
import subprocess
import sys
from pathlib import Path

from typer.testing import CliRunner

from maaspower import __version__
from maaspower.__main__ import cli

runner = CliRunner()


def test_cli_version():
    cmd = [sys.executable, "-m", "maaspower", "--version"]
    assert subprocess.check_output(cmd).decode().strip() == __version__


def test_pmac_schema(tmp_path: Path, samples: Path):
    """generate the schema file for maaspower configuration YAML"""
    schema_path = tmp_path / "maaspower.schema.json"
    result = runner.invoke(cli, ["schema", str(schema_path)])
    assert result.exit_code == 0, f"schema failed with: {result}"

    with open(samples / "maaspower.schema.json") as stream:
        expected = json.loads(stream.read())

    with open(schema_path) as stream:
        actual = json.loads(stream.read())

    assert expected == actual



"""
kasa_device.py
--------------

This module contains the class to represent and control TP-Link Kasa smart devices via
the python-kasa API. It provides functionalities to turn the device on or off
and to query its current state.
"""

import asyncio
from dataclasses import dataclass
from typing import Annotated as A
from typing import Literal

from kasa import SmartDeviceException, SmartPlug

from maaspower.maas_globals import desc
from maaspower.maasconfig import SwitchDevice


@dataclass(kw_only=True)
class KasaDevice(SwitchDevice):
    ip_address: A[str, desc("The IP address of the Kasa device.")]
    type: Literal["KasaDevice"] = "KasaDevice"
    name: A[str, desc("The name of the switching device.")]

    def __post_init__(self):
        """Initialize the smart plug with the provided IP address
        after the object is instantiated."""
        super().__post_init__()
        # self.plug = SmartPlug(self.ip_address)

    def turn_on(self) -> None:
        """Asynchronously turn the smart plug on using a coroutine
        managed by asyncio."""
        asyncio.run(self._turn_on())

    async def _turn_on(self):
        """Private coroutine to turn the smart plug on, catching
        and logging any exceptions."""
        plug = SmartPlug(self.ip_address)
        try:
            await plug.update()
            await plug.turn_on()
        except SmartDeviceException as e:
            print(f"Failed to turn on the smart plug at {self.ip_address}: {str(e)}")

    def turn_off(self) -> None:
        """Asynchronously turn the smart plug off using a coroutine
        managed by asyncio."""
        asyncio.run(self._turn_off())

    async def _turn_off(self):
        """Private coroutine to turn the smart plug off, catching
        and logging any exceptions."""
        plug = SmartPlug(self.ip_address)
        try:
            await plug.update()
            await plug.turn_off()
        except SmartDeviceException as e:
            print(f"Failed to turn off the smart plug at {self.ip_address}: {str(e)}")

    def query_state(self) -> str:
        """Asynchronously query the current state of the smart plug
        and return it as 'on' or 'off'."""
        return asyncio.run(self._query_state())

    async def _query_state(self) -> str:
        """Private coroutine to query the state of the smart plug,
        returning 'on', 'off', or 'error' if an exception occurs."""
        plug = SmartPlug(self.ip_address)
        try:
            await plug.update()
            return "on" if plug.is_on else "off"
        except SmartDeviceException as e:
            print(
                f"Failed to query state of the smart plug \
                    at {self.ip_address}: {str(e)}"
            )
            return "error"



"""
maaspower package
=================
"""

"""Top level API.

.. data:: __version__
    :type: str

    Version number as calculated by https://github.com/pypa/setuptools_scm
"""

from ._version import __version__  # noqa: E402

# __all__ defines the public API for the package.
# Each module also defines its own __all__.
__all__ = ["__version__", "maasconfig", "webhook"]



# Note get the chrome driver for selenium from here:
# https://chromedriver.storage.googleapis.com/index.html?path=100.0.4896.20/

import re
from dataclasses import dataclass
from enum import Enum
from time import sleep
from typing import Annotated as A
from typing import Literal

from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

from maaspower.maas_globals import desc
from maaspower.maasconfig import RegexSwitchDevice

command_regex = re.compile(r"([^\/]*)\/([^\/]*)\/?([^\/]*)?\/?([^\/]*)?$")
index_regex = re.compile(r"(.*)\[([0-9]*)\]")


class FindBy(Enum):
    id = By.ID
    cls = By.CLASS_NAME
    link = By.PARTIAL_LINK_TEXT
    n = By.NAME
    css = By.CSS_SELECTOR


@dataclass(kw_only=True)
class WebGui(RegexSwitchDevice):
    """A device controlled via a Web GUI"""

    type: Literal["WebGui"] = "WebGui"

    connect_url: A[str, desc("URL to Web UI")] = "none"
    login: A[str, desc("command sequence to log in")] = "none"
    logout: A[str, desc("command sequence to log out")] = "none"
    timeout: A[int, desc("max timeout on any UI transitions")] = 10
    driver: A[str, desc("Path to chromedriver binary")] = "./chromedriver"

    # this gets called after the dataclass __init__
    def __post_init__(self):
        self.last_get = ""
        self.connect()

    def __del__(self):
        self.disconnect()

    def connect(self, retries=2):
        options = webdriver.ChromeOptions()
        options.add_argument("headless")
        self.c_driver = webdriver.Chrome(self.driver, options=options)

        self.c_driver.get(self.connect_url)
        self.c_driver.timeouts._implicit_wait = self.timeout

        self.execute_command(self.login, retries=retries)

    def disconnect(self):
        self.execute_command(self.logout, retries=0)
        try:
            self.c_driver.close()
        except Exception:
            pass

    def execute_command(self, command_list: str, retries=2):
        """
        Take a sequence of command strings separated by \n.
        Each command identifies and interacts with an HTML element
        """

        element_list = command_list.split("\n")[:-1]
        for element in element_list:
            # separate out the elements of the string
            match = command_regex.match(element)
            if match is None:
                raise ValueError(f"bad command format: {element} {element_list}")

            command = match.group(1)

            while retries >= 0:
                try:
                    # invoke the command via selenium
                    if command == "click":
                        self.click(match.group(2), match.group(3))
                    elif command == "send":
                        self.send(match.group(2), match.group(3), match.group(4))
                    elif command == "sendcr":
                        self.send(
                            match.group(2), match.group(3), match.group(4), cr=True
                        )
                    elif command == "get":
                        self.last_get = self.get(match.group(2), match.group(3))
                    elif command == "delay":
                        sleep(float(match.group(2)))
                except Exception:
                    retries -= 1
                    if retries > 0:
                        self.disconnect()
                        self.connect(retries=0)
                else:
                    # success - leave the retry loop
                    break
            else:
                # abort remaining commands when failed retry times
                return

    def process_arguments(self, by_str: str, value: str) -> tuple[str, str, int | None]:
        # convert our short fieldtype string to the full selenium string
        by = FindBy[by_str].value

        # check for and index [n] at the end of a string. Used for
        # matching > 1 html element and we want to index into the list of matches
        idx_match = index_regex.match(value)
        if idx_match is None:
            index = None
        else:
            value = idx_match.group(1)
            index = int(idx_match.group(2))
        return by, value, index

    def click(self, by_str: str, value: str):
        by, value, index = self.process_arguments(by_str, value)
        elem = WebDriverWait(self.c_driver, self.timeout).until(
            EC.element_to_be_clickable((by, value))
        )
        if index is not None:
            elements = self.c_driver.find_elements(by=by, value=value)
            elem = elements[index]
        elem.click()

    def send(self, by_str: str, value: str, text: str, cr=False):
        by, value, index = self.process_arguments(by_str, value)
        if cr:
            text += "\n"
        elem = WebDriverWait(self.c_driver, self.timeout).until(
            EC.presence_of_element_located((by, value))
        )
        if index:
            elements = self.c_driver.find_elements(by=by, value=value)
            elem = elements[index]
        elem.send_keys(text)

    def get(self, by_str: str, value) -> str:
        by, value, index = self.process_arguments(by_str, value)
        elem = WebDriverWait(self.c_driver, self.timeout).until(
            EC.presence_of_element_located((by, value))
        )
        if index:
            elements = self.c_driver.find_elements(by=by, value=value)
            elem = elements[index]
        return elem.text
